package com.infinispan.hotrod.client;

import java.net.URL;
import java.text.DecimalFormat;
import java.util.Date;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.ExecutionException;

import org.infinispan.client.hotrod.RemoteCache;
import org.infinispan.client.hotrod.RemoteCacheManager;
import org.infinispan.client.hotrod.ServerStatistics;
import org.infinispan.util.concurrent.NotifyingFuture;

import com.infinispan.hotrod.domain.Person;



public class HotRodClient {

	private static final String WORKLOAD_PROPERTIES       				= "workload.properties";
	private static final String HOTROD_CLIENT_PROPERTIES  				= "hotrod-client.properties";

	private static final String CACHE_NAME                				= "cache.name";
	private static final String HOTROD_CLIENT_PROPERTIES_FILE           = "hotrod.client.properties.file";
	private static final String TOTAL_TRANSACTION_COUNT        			= "total.transaction.count";
	private static final String READ_PROPORTION                     	= "read.proportion";
	private static final String WRITE_PROPORTION                    	= "write.proportion";
	private static final String READ_COMMUNICATION_MODE                	= "read.communication.mode";
	private static final String WRITE_COMMUNICATION_MODE                = "write.communication.mode";
	private static final String OBJECT_SIZE_DISTRIBUTION 				= "object.size.distribution";
	private static final String TRANSACTION_PROPORTION_DISTRIBUTION 	= "transation.proportion.distribution";

	private static final String DEFAULT_NAME              				= "default";

	private static final DecimalFormat df = new DecimalFormat("#.##");

	private RemoteCache<String, Person> cache = null;

	private String clientName = "0";
	private String cacheName;
	private String hotrodClientProperties;
	private int readTranCount;
	private int writeTranCount;
	private String readCommMode;
	private String writeCommMode;
	private String[] objType;
	private int[] objSize;
	private int[] readTranCountPerObjType;
	private int[] writeTranCountPerObjType;
	private double[] msPerTranWritePerObjType;
	private double[] msPerTranReadPerObjType;
	private double[] tranPerSecWritePerObjType;
	private double[] tranPerSecReadPerObjType;
	private long[] totalReadTimePerObjType;
	private long[] totalWriteTimePerObjType;

	private long totalWriteTime;
	private long totalReadTime;
	private double msPerTranWrite;
	private double msPerTranRead;
	private double tranPerSecWrite;
	private double tranPerSecRead;

	private String startTime;
	private String endTime;

	/**
	 * 
	 * @throws InterruptedException
	 * @throws ExecutionException
	 */
	private void execute() throws InterruptedException, ExecutionException {
		startTime = new Date().toString();

		//		RemoteCacheManager rcm = new RemoteCacheManager("127.0.0.1:11223");

		URL resource = Thread.currentThread().getContextClassLoader().getResource(hotrodClientProperties);
		RemoteCacheManager rcm = new RemoteCacheManager(resource, true);

		if (cacheName.equals(DEFAULT_NAME)) {
			cache = rcm.getCache();
		} else {
			cache = rcm.getCache(cacheName);
		}

		if (writeTranCount > 0) {
			writeToCache();
		}

		if (readTranCount > 0) {
			readFromCache();
		}
		endTime = new Date().toString();

		Map<String, String> statMap = getStatistics();
		printOutput(statMap);

	}

	/**
	 * 
	 */
	private void writeToCache() {
		msPerTranWritePerObjType = new double[objSize.length];
		tranPerSecWritePerObjType = new double[objSize.length];
		totalWriteTimePerObjType = new long[objSize.length];

		for (int i = 0; i < objSize.length; i++) {
			long writeTime = 0;
			for (int j = 0; j < writeTranCountPerObjType[i]; j++) {
				String key = "name"+clientName+i+j;               
				Person p = new Person(key, objSize[i]);

				long timeDifference = 0;
				if (writeCommMode.equalsIgnoreCase(CommunicationMode.sync.name())) {
					timeDifference = writeSync(key, p);
				} else {
					timeDifference = writeAsync(key, p);
				}				

				writeTime += timeDifference;
			}
			totalWriteTime += writeTime;
			totalWriteTimePerObjType[i] = writeTime;
			msPerTranWritePerObjType[i] = (double) writeTime / writeTranCountPerObjType[i];
			tranPerSecWritePerObjType[i] = (double) (writeTranCountPerObjType[i] * 1000) / writeTime;
		} 

		msPerTranWrite = (double) totalWriteTime / writeTranCount;
		tranPerSecWrite = (double) (writeTranCount * 1000) / totalWriteTime;
	} 

	/**
	 * 
	 * @throws InterruptedException
	 * @throws ExecutionException
	 */
	private void readFromCache() throws InterruptedException, ExecutionException {	
		msPerTranReadPerObjType = new double[objSize.length];
		tranPerSecReadPerObjType = new double[objSize.length];
		totalReadTimePerObjType = new long[objSize.length];

		for (int i = 0; i < objSize.length; i++) {
			long readTime = 0;
			for (int j = 0; j < (readTranCountPerObjType[i]); j++) {
				String key = "name"+clientName+i+j; 

				long timeDifference = 0;
				if (readCommMode.equalsIgnoreCase(CommunicationMode.sync.name())) {
					timeDifference = readSync(key);
				} else {
					timeDifference = readAsync(key);
				}

				readTime += timeDifference;
			}
			totalReadTime += readTime;
			totalReadTimePerObjType[i] = readTime;
			msPerTranReadPerObjType[i] = (double) readTime / readTranCountPerObjType[i];
			tranPerSecReadPerObjType[i] = (double) (readTranCountPerObjType[i] * 1000) / readTime;
		}	

		msPerTranRead = (double) totalReadTime / readTranCount;
		tranPerSecRead = (double) (readTranCount * 1000) / totalReadTime;
	}


	/**
	 * 
	 * @param key
	 * @param p
	 * @return
	 */
	private long writeSync(String key, Person p) {
		long startTime = new Date().getTime(); 

		cache.put(key, p);            

		long endTime = new Date().getTime();   
		return endTime - startTime; 
	}

	/**
	 * 
	 * @param key
	 * @param p
	 * @return
	 */
	private long writeAsync(String key, Person p) {
		long startTime = new Date().getTime(); 

		cache.putAsync(key, p);            

		long endTime = new Date().getTime();   
		return endTime - startTime; 
	}

	/**
	 * 
	 * @param key
	 * @return
	 */
	private long readSync(String key) {
		long startTime = new Date().getTime();  

		Person p = (Person) cache.get(key);

		//		VersionedValue<Person> p = (VersionedValue<Person>) rc.getVersioned(key);

		long endTime = new Date().getTime(); 

		if (p == null) {
			System.out.println("No object found for key = " + key);
		}

		return endTime - startTime; 
	}

	/**
	 * 
	 * @param key
	 * @return
	 * @throws InterruptedException
	 * @throws ExecutionException
	 */
	private long readAsync(String key) throws InterruptedException, ExecutionException {
		long startTime = new Date().getTime(); 

		NotifyingFuture<Person> nf = (NotifyingFuture<Person>) cache.getAsync(key);
		//      Person p = nf.get();

		long endTime = new Date().getTime();  

		//		Thread.sleep(40);
		//		System.out.println(nf.isDone() + " for key = " + key);

		//		if (p == null) {
		//			System.out.println("No object found for key = " + key);
		//		}

		return endTime - startTime; 
	}

	/**
	 * 
	 * @return
	 */
	private Map<String, String> getStatistics() {
		ServerStatistics serverStats = cache.stats();
		Map<String, String> statMap = serverStats.getStatsMap();
		return statMap;
	}

	/**
	 * 
	 * @param statMap
	 */
	private void printOutput(Map<String, String> statMap) {
		String outStr = "=================== Client Name = Client " + clientName + ", Start time = " + startTime  + " ===================";
		outStr += "\nCache Name		= " + cache.getName();
		outStr += "\nTotal Write Count	= " + writeTranCount + ", Communication mode = " + writeCommMode;
		outStr += "\nTotal Read Count	= " + readTranCount + ", Communication mode = " + readCommMode;
		if (writeTranCount > 0) {
			outStr += "\n";
			for (int i = 0; i < objSize.length; i++) {
				outStr += "\n(Write) For Object Size: " + objType[i] + ", 	Total Txn Count = " + writeTranCountPerObjType[i] + ", 	Total Time (ms) = " + totalWriteTimePerObjType[i] + ", 	Avg Time (ms) = " + df.format(msPerTranWritePerObjType[i]) + ", 	Txns Per Sec = " + df.format(tranPerSecWritePerObjType[i]);
			}
			outStr += "\n(Write) For All Objects: 	Total Txn Count = " + writeTranCount + ", 	Total Time (ms) = " + totalWriteTime + ", 	Avg Time (ms) = " + df.format(msPerTranWrite) + ", 	Txns per sec = " + df.format(tranPerSecWrite);			
		}
		if (readTranCount > 0) {
			outStr += "\n";
			for (int i = 0; i < objSize.length; i++) {
				outStr += "\n(Read) For Object Size: " + objType[i] + ", 	Total Txn Count = " + readTranCountPerObjType[i] + ", 	Total Time (ms) = " + totalReadTimePerObjType[i] + ", 	Avg Time (ms) = " + df.format(msPerTranReadPerObjType[i]) + ", 	Txns Per Sec = " + df.format(tranPerSecReadPerObjType[i]);
			}
			outStr += "\n(Read) For All Objects: 	Total Txn Count = " + readTranCount + ", 	Total Time (ms) = " + totalReadTime + ", 	Avg Time (ms) = " + df.format(msPerTranRead) + ", 	Txns per sec = " + df.format(tranPerSecRead);
		}
		outStr += "\n\nStatistics :";
		Iterator<Map.Entry<String, String>> it = statMap.entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry<String, String> stat = it.next();
			outStr += "\n" + stat.getKey() + " = " + stat.getValue();
		}

		outStr += "\n=================== Client Name = Client " + clientName + ", End time   = " + endTime  + " ===================";
		System.out.println(outStr);
	}

	/**
	 * 
	 * @param propertyFile
	 * @throws Exception
	 */
	private void loadProperties(String propertyFile) throws Exception {
		Properties p = new Properties();

		p.load(Thread.currentThread().getContextClassLoader().getResourceAsStream(propertyFile));
		cacheName = p.getProperty(CACHE_NAME, DEFAULT_NAME).trim();
		hotrodClientProperties = p.getProperty(HOTROD_CLIENT_PROPERTIES_FILE, HOTROD_CLIENT_PROPERTIES).trim();
		int tranCount = Integer.parseInt(p.getProperty(TOTAL_TRANSACTION_COUNT, "200").trim());
		float readProportion = Float.parseFloat(p.getProperty(READ_PROPORTION, "0.5").trim());
		float writeProportion = Float.parseFloat(p.getProperty(WRITE_PROPORTION, "0.5").trim());
		objType = p.getProperty(OBJECT_SIZE_DISTRIBUTION, "1KB,1MB").trim().split(",");
		String[] tranDistArr = p.getProperty(TRANSACTION_PROPORTION_DISTRIBUTION, "0.8,0.2").trim().split(",");

		readCommMode = p.getProperty(READ_COMMUNICATION_MODE, "sync").trim();
		writeCommMode = p.getProperty(WRITE_COMMUNICATION_MODE, "sync").trim();

		readTranCount = (int) (tranCount * readProportion);
		writeTranCount = (int) (tranCount * writeProportion);

		objSize = new int[objType.length];
		for (int i = 0; i < objType.length; i++) {
			String str = objType[i].trim();
			int num = Integer.parseInt(str.substring(0, str.length()-2));
			String type = str.substring(str.length()-2).toUpperCase();
			int bytes = ObjectSize.valueOf(type).getSize();
			objSize[i] = num * bytes;
		}

		readTranCountPerObjType = new int[tranDistArr.length];
		writeTranCountPerObjType = new int[tranDistArr.length];
		for (int i = 0; i < tranDistArr.length; i++) {
			String str = tranDistArr[i].trim();
			float num = Float.parseFloat(str);
			readTranCountPerObjType[i] = (int) (num * readTranCount);
			writeTranCountPerObjType[i] = (int) (num * writeTranCount);
		}
	}

	/**
	 * 
	 * @param args
	 */
	public HotRodClient(String[] args) {
		String propertyFile = null;
		if (args.length > 0) {
			propertyFile = args[0];
		} else {
			propertyFile = WORKLOAD_PROPERTIES;
		}
		try {
			loadProperties(propertyFile);
		} catch (Exception e) {
			System.out.println("Loading failed...");
			e.printStackTrace();
		}
		if (args.length > 1) {
			clientName = args[1];
		}
		try {
			execute();
		} catch (InterruptedException e) {
			e.printStackTrace();
		} catch (ExecutionException e) {
			e.printStackTrace();
		}
	}


	/**
	 * 
	 * @param args
	 * @throws Exception
	 */
	public static void main(String[] args) throws Exception {
		new HotRodClient(args);
	}

	/**
	 * 
	 *
	 */
	private enum ObjectSize {
		KB(1024), MB(1048576);

		private int size;

		private ObjectSize(int size) {
			this.size = size;
		}

		public int getSize() {
			return size;
		}
	}

	/**
	 * 
	 *
	 */
	private enum CommunicationMode {
		sync, async;
	}
}
